@@ -23,6 +23,7 @@
  */
 package hudson.remoting;
 
+import hudson.remoting.CommandTransport.CommandReceiver;
 import hudson.remoting.ExportTable.ExportList;
 import hudson.remoting.PipeWindow.Key;
 import hudson.remoting.PipeWindow.Real;
@@ -30,10 +31,8 @@ import hudson.remoting.forward.ListeningPort;
 import hudson.remoting.forward.ForwarderFactory;
 import hudson.remoting.forward.PortForwarder;
 
-import java.io.EOFException;
 import java.io.IOException;
 import java.io.InputStream;
-import java.io.ObjectInputStream;
 import java.io.ObjectOutputStream;
 import java.io.OutputStream;
 import java.io.PrintWriter;
@@ -107,10 +106,13 @@ import java.util.concurrent.ThreadFactory;
  * @author Kohsuke Kawaguchi
  */
 public class Channel implements VirtualChannel, IChannel {
-    private final ObjectInputStream ois;
-    private final ObjectOutputStream oos;
+    private final CommandTransport transport;
+
     /**
      * {@link OutputStream} that's given to the constructor. This is the hand-off with the lower layer.
+     *
+     * @deprecated
+     *      See {@link #getUnderlyingOutput()}.
      */
     private final OutputStream underlyingOutput;
 
@@ -120,7 +122,7 @@ public class Channel implements VirtualChannel, IChannel {
      */
     private final String name;
     private volatile boolean isRestricted;
-    /*package*/ final ExecutorService executor;
+    /*package*/ final InterceptingExecutorService executor;
 
     /**
      * If non-null, the incoming link is already shut down,
@@ -135,8 +137,23 @@ public class Channel implements VirtualChannel, IChannel {
      */
     private volatile Throwable outClosed = null;
 
+    /**
+     * Requests that are sent to the remote side for execution, yet we are waiting locally until
+     * we hear back their responses.
+     */
     /*package*/ final Map<Integer,Request<?,?>> pendingCalls = new Hashtable<Integer,Request<?,?>>();
 
+    /**
+     * Remembers last I/O ID issued from locally to the other side, per thread.
+     * int[1] is used as a holder of int.
+     */
+    private final ThreadLocal<int[]> lastIoId = new ThreadLocal<int[]>() {
+        @Override
+        protected int[] initialValue() {
+            return new int[1];
+        }
+    };
+
     /**
      * Records the {@link Request}s being executed on this channel, sent by the remote peer.
      */
@@ -213,6 +230,8 @@ public class Channel implements VirtualChannel, IChannel {
      */
     public final AtomicInteger resourceLoadingCount = new AtomicInteger();
 
+    private final AtomicInteger ioId = new AtomicInteger();
+
     /**
      * Property bag that contains application-specific stuff.
      */
@@ -221,7 +240,7 @@ public class Channel implements VirtualChannel, IChannel {
     /**
      * Proxy to the remote {@link Channel} object.
      */
-    private IChannel remoteChannel;
+    private final IChannel remoteChannel;
 
     /**
      * Capability of the remote {@link Channel}.
@@ -248,7 +267,7 @@ public class Channel implements VirtualChannel, IChannel {
      * in case read/write blocks. It is single thread to ensure FIFO; I/O needs to execute
      * in the same order the remote peer told us to execute them.
      */
-    /*package*/ final ExecutorService pipeWriter;
+    /*package*/ final PipeWriter pipeWriter;
 
     /**
      * ClassLaoder that remote classloaders should use as the basis.
@@ -256,7 +275,8 @@ public class Channel implements VirtualChannel, IChannel {
     /*package*/ final ClassLoader baseClassLoader;
 
     /**
-     * Communication mode.
+     * Communication mode used in conjunction with {@link ClassicCommandTransport}.
+     * 
      * @since 1.161
      */
     public enum Mode {
@@ -289,7 +309,7 @@ public class Channel implements VirtualChannel, IChannel {
          * if the preamble is "AAB", we'll fail to find a preamble
          * in "AAAB".
          */
-        private final byte[] preamble;
+        /*package*/ final byte[] preamble;
 
         Mode(String preamble) {
             try {
@@ -367,10 +387,29 @@ public class Channel implements VirtualChannel, IChannel {
     }
 
     /*package*/ Channel(String name, ExecutorService exec, Mode mode, InputStream is, OutputStream os, OutputStream header, boolean restricted, ClassLoader base, Capability capability) throws IOException {
+        this(name,exec, ClassicCommandTransport.create(mode, is, os, header, base, capability),restricted,base);
+    }
+
+    /**
+     * Creates a new channel.
+     *
+     * @param name
+     *      See {@link #Channel(String, ExecutorService, Mode, InputStream, OutputStream, OutputStream, boolean, ClassLoader)}
+     * @param exec
+     *      See {@link #Channel(String, ExecutorService, Mode, InputStream, OutputStream, OutputStream, boolean, ClassLoader)}
+     * @param transport
+     *      The transport that we run {@link Channel} on top of.
+     * @param base
+     *      See {@link #Channel(String, ExecutorService, Mode, InputStream, OutputStream, OutputStream, boolean, ClassLoader)}
+     * @param restricted
+     *      See {@link #Channel(String, ExecutorService, Mode, InputStream, OutputStream, OutputStream, boolean, ClassLoader)}
+     * @since 2.13
+     */
+    public Channel(String name, ExecutorService exec, CommandTransport transport, boolean restricted, ClassLoader base) throws IOException {
         this.name = name;
-        this.executor = exec;
+        this.executor = new InterceptingExecutorService(exec);
         this.isRestricted = restricted;
-        this.underlyingOutput = os;
+        this.underlyingOutput = transport.getUnderlyingStream();
 
         if (base==null)
             base = getClass().getClassLoader();
@@ -380,74 +419,28 @@ public class Channel implements VirtualChannel, IChannel {
             throw new AssertionError(); // export number 1 is reserved for the channel itself
         remoteChannel = RemoteInvocationHandler.wrap(this,1,IChannel.class,true,false);
 
-        // write the magic preamble.
-        // certain communication channel, such as forking JVM via ssh,
-        // may produce some garbage at the beginning (for example a remote machine
-        // might print some warning before the program starts outputting its own data.)
-        //
-        // so use magic preamble and discard all the data up to that to improve robustness.
+        this.remoteCapability = transport.getRemoteCapability();
+        this.pipeWriter = new PipeWriter(createPipeWriterExecutor());
 
-        capability.writePreamble(os);
+        this.transport = transport;
 
-        ObjectOutputStream oos = null;
-        if(mode!= Mode.NEGOTIATE) {
-            os.write(mode.preamble);
-            oos = new ObjectOutputStream(mode.wrap(os));
-            oos.flush();    // make sure that stream preamble is sent to the other end. avoids dead-lock
-        }
-
-        {// read the input until we hit preamble
-            Mode[] modes={Mode.BINARY,Mode.TEXT};
-            byte[][] preambles = new byte[][]{Mode.BINARY.preamble, Mode.TEXT.preamble, Capability.PREAMBLE};
-            int[] ptr=new int[3];
-            Capability cap = new Capability(0); // remote capacity that we obtained. If we don't hear from remote, assume no capability
-
-            while(true) {
-                int ch = is.read();
-                if(ch==-1)
-                    throw new EOFException("unexpected stream termination");
-
-                for(int i=0;i<preambles.length;i++) {
-                    byte[] preamble = preambles[i];
-                    if(preamble[ptr[i]]==ch) {
-                        if(++ptr[i]==preamble.length) {
-                            switch (i) {
-                            case 0:
-                            case 1:
-                                // transmission mode negotiation
-                                if(mode==Mode.NEGOTIATE) {
-                                    // now we know what the other side wants, so send the consistent preamble
-                                    mode = modes[i];
-                                    os.write(mode.preamble);
-                                    oos = new ObjectOutputStream(mode.wrap(os));
-                                    oos.flush();
-                                } else {
-                                    if(modes[i]!=mode)
-                                        throw new IOException("Protocol negotiation failure");
-                                }
-                                this.oos = oos;
-                                this.remoteCapability = cap;
-                                this.pipeWriter = createPipeWriter();
-                                this.ois = new ObjectInputStreamEx(mode.wrap(is),base);
-                                new ReaderThread(name).start();
-
-                                return;
-                            case 2:
-                                cap = Capability.read(is);
-                                break;
-                            }
-                            ptr[i]=0; // reset
-                        }
-                    } else {
-                        // didn't match.
-                        ptr[i]=0;
-                    }
+        transport.setup(this, new CommandReceiver() {
+            public void handle(Command cmd) {
+                lastHeard = System.currentTimeMillis();
+                if (logger.isLoggable(Level.FINE))
+                    logger.fine("Received " + cmd);
+                try {
+                    cmd.execute(Channel.this);
+                } catch (Throwable t) {
+                    logger.log(Level.SEVERE, "Failed to execute command " + cmd + " (channel " + Channel.this.name + ")", t);
+                    logger.log(Level.SEVERE, "This command is created here", cmd.createdAt);
                 }
-
-                if(header!=null)
-                    header.write(ch);
             }
-        }
+
+            public void terminate(IOException e) {
+                Channel.this.terminate(e);
+            }
+        });
     }
 
     /**
@@ -477,7 +470,7 @@ public class Channel implements VirtualChannel, IChannel {
      * reader thread (thus prevent blockage.) Otherwise let the channel reader thread do it,
      * which is the historical behaviour.
      */
-    private ExecutorService createPipeWriter() {
+    private ExecutorService createPipeWriterExecutor() {
         if (remoteCapability.supportsPipeThrottling())
             return Executors.newSingleThreadExecutor(new ThreadFactory() {
                 public Thread newThread(Runnable r) {
@@ -499,20 +492,9 @@ public class Channel implements VirtualChannel, IChannel {
             throw new ChannelClosedException(outClosed);
         if(logger.isLoggable(Level.FINE))
             logger.fine("Send "+cmd);
-        Channel old = Channel.setCurrent(this);
-        try {
-            oos.writeObject(cmd);
-            oos.flush();        // make sure the command reaches the other end.
-        } finally {
-            Channel.setCurrent(old);
-            commandsSent++;
-        }
-        // unless this is the last command, have OOS and remote OIS forget all the objects we sent
-        // in this command. Otherwise it'll keep objects in memory unnecessarily.
-        // However, this may fail if the command was the close, because that's supposed to be the last command
-        // ever sent. See the comment from jglick on HUDSON-3077 about what happens if we do oos.reset(). 
-        if(!(cmd instanceof CloseCommand))
-            oos.reset();
+
+        transport.write(cmd, cmd instanceof CloseCommand);
+        commandsSent++;
     }
 
     /**
@@ -726,6 +708,11 @@ public class Channel implements VirtualChannel, IChannel {
     protected synchronized void terminate(IOException e) {
         if (e==null)    throw new IllegalArgumentException();
         outClosed=inClosed=e;
+        try {
+            transport.closeRead();
+        } catch (IOException x) {
+            logger.log(Level.WARNING, "Failed to close down the reader side of the transport",x);
+        }
         try {
             synchronized(pendingCalls) {
                 for (Request<?,?> req : pendingCalls.values())
@@ -767,6 +754,24 @@ public class Channel implements VirtualChannel, IChannel {
         return listeners.remove(l);
     }
 
+    /**
+     * Adds a {@link CallableFilter} that gets a chance to decorate every {@link Callable}s that run locally
+     * sent by the other peer.
+     * 
+     * This is useful to tweak the environment those closures are run, such as setting up the thread context
+     * environment.
+     */
+    public void addLocalExecutionInterceptor(CallableFilter filter) {
+        executor.addFilter(filter);
+    }
+
+    /**
+     * Rmoves the filter introduced by {@link #addLocalExecutionInterceptor(CallableFilter)}.
+     */
+    public void removeLocalExecutionInterceptor(CallableFilter filter) {
+        executor.removeFilter(filter);
+    }
+    
     /**
      * Waits for this {@link Channel} to be closed down.
      *
@@ -815,7 +820,7 @@ public class Channel implements VirtualChannel, IChannel {
     /**
      * Notifies the remote peer that we are closing down.
      *
-     * Execution of this command also triggers the {@link ReaderThread} to shut down
+     * Execution of this command also triggers the {@link SynchronousCommandTransport.ReaderThread} to shut down
      * and quit. The {@link CloseCommand} is always the last command to be sent on
      * {@link ObjectOutputStream}, and it's the last command to be read.
      */
@@ -889,7 +894,7 @@ public class Channel implements VirtualChannel, IChannel {
         send(new CloseCommand(diagnosis));
         outClosed = new IOException().initCause(diagnosis);   // last command sent. no further command allowed. lock guarantees that no command will slip inbetween
         try {
-            oos.close();
+            transport.closeWrite();
         } catch (IOException e) {
             // there's a race condition here.
             // the remote peer might have already responded to the close command
@@ -1036,6 +1041,28 @@ public class Channel implements VirtualChannel, IChannel {
                 ForwarderFactory.create(forwardHost, forwardPort));
     }
 
+    /**
+     * Dispenses the unique I/O ID.
+     *
+     * When a {@link Channel} requests an activity that happens in {@link #pipeWriter},
+     * the sender assigns unique I/O ID to this request, which enables later
+     * commands to sync up with their executions.
+     *
+     * @see PipeWriter
+     */
+    /*package*/ int newIoId() {
+        int v = ioId.incrementAndGet();
+        lastIoId.get()[0] = v;
+        return v;
+    }
+
+    /**
+     * Gets the last I/O ID issued by the calling thread, or 0 if none is recorded.
+     */
+    /*package*/ int lastIoId() {
+        return lastIoId.get()[0];
+    }
+
     /**
      * Blocks until all the I/O packets sent before this gets fully executed by the remote side, then return.
      *
@@ -1046,15 +1073,34 @@ public class Channel implements VirtualChannel, IChannel {
         call(new IOSyncer());
     }
 
+//  Barrier doesn't work because IOSyncer is a Callable and not Command
+//  (yet making it Command would break JENKINS-5977, which introduced this split in the first place!)
+//    /**
+//     * Non-blocking version of {@link #syncIO()} that has a weaker commitment.
+//     *
+//     * This method only guarantees that any later remote commands will happen after all the I/O packets sent before
+//     * this method call gets fully executed. This is faster in that it it doesn't wait for a response
+//     * from the other side, yet it normally achieves the desired semantics.
+//     */
+//    public void barrierIO() throws IOException {
+//        callAsync(new IOSyncer());
+//    }
+
     public void syncLocalIO() throws InterruptedException {
+        Thread t = Thread.currentThread();
+        String old = t.getName();
+        t.setName("I/O sync: "+old);
         try {
-            pipeWriter.submit(new Runnable() {
+            // no one waits for the completion of this Runnable, so not using I/O ID
+            pipeWriter.submit(0,new Runnable() {
                 public void run() {
                     // noop
                 }
             }).get();
         } catch (ExecutionException e) {
             throw new AssertionError(e); // impossible
+        } finally {
+            t.setName(old);
         }
     }
 
@@ -1067,6 +1113,10 @@ public class Channel implements VirtualChannel, IChannel {
         private static final long serialVersionUID = 1L;
     }
 
+    public String getName() {
+        return name;
+    }
+
     @Override
     public String toString() {
         return super.toString()+":"+name;
@@ -1090,67 +1140,6 @@ public class Channel implements VirtualChannel, IChannel {
         return lastHeard;
     }
 
-    private final class ReaderThread extends Thread {
-        private int commandsReceived = 0;
-        private int commandsExecuted = 0;
-
-        public ReaderThread(String name) {
-            super("Channel reader thread: "+name);
-        }
-
-        @Override
-        public void run() {
-            try {
-                while(inClosed==null) {
-                    Command cmd = null;
-                    try {
-                        Channel old = Channel.setCurrent(Channel.this);
-                        try {
-                            cmd = (Command)ois.readObject();
-                            lastHeard = System.currentTimeMillis();
-                        } finally {
-                            Channel.setCurrent(old);
-                        }
-                    } catch (EOFException e) {
-                        IOException ioe = new IOException("Unexpected termination of the channel");
-                        ioe.initCause(e);
-                        throw ioe;
-                    } catch (ClassNotFoundException e) {
-                        logger.log(Level.SEVERE, "Unable to read a command (channel " + name + ")",e);
-                        continue;
-                    } finally {
-                        commandsReceived++;
-                    }
-
-                    if(logger.isLoggable(Level.FINE))
-                        logger.fine("Received "+cmd);
-                    try {
-                        cmd.execute(Channel.this);
-                    } catch (Throwable t) {
-                        logger.log(Level.SEVERE, "Failed to execute command "+cmd+ " (channel " + name + ")",t);
-                        logger.log(Level.SEVERE, "This command is created here",cmd.createdAt);
-                    } finally {
-                        commandsExecuted++;
-                    }
-                }
-                ois.close();
-            } catch (IOException e) {
-                logger.log(Level.SEVERE, "I/O error in channel "+name,e);
-                terminate(e);
-            } catch (RuntimeException e) {
-                logger.log(Level.SEVERE, "Unexpected error in channel "+name,e);
-                terminate((IOException)new IOException("Unexpected reader termination").initCause(e));
-                throw e;
-            } catch (Error e) {
-                logger.log(Level.SEVERE, "Unexpected error in channel "+name,e);
-                terminate((IOException)new IOException("Unexpected reader termination").initCause(e));
-                throw e;
-            } finally {
-                pipeWriter.shutdown();
-            }
-        }
-    }
-
     /*package*/ static Channel setCurrent(Channel channel) {
         Channel old = CURRENT.get();
         CURRENT.set(channel);
@@ -1176,7 +1165,7 @@ public class Channel implements VirtualChannel, IChannel {
 
     private static final Logger logger = Logger.getLogger(Channel.class.getName());
 
-    public static final int PIPE_WINDOW_SIZE = Integer.getInteger(Channel.class+".pipeWindowSize",128*1024);
+    public static final int PIPE_WINDOW_SIZE = Integer.getInteger(Channel.class.getName()+".pipeWindowSize",128*1024);
 
 //    static {
 //        ConsoleHandler h = new ConsoleHandler();
